import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="451",
    message="Child is still freed for adoption, but freeing orders could not be applied for since 30 December 2005.",
    affected_fields=["DEC", "REC", "LS"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        mask = (
            episodes["DEC"].isna() & episodes["REC"].isna() & (episodes["LS"] == "D1")
        )
        error_locations = episodes.index[mask]
        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103", "104"],
            "REC": ["E11", "E12", "X1", pd.NA, "E11", "E11"],
            "DEC": [
                "15/03/2021",
                "17/06/2020",
                "20/03/2020",
                pd.NA,
                "23/08/2020",
                pd.NA,
            ],
            "LS": ["J2", "X", "J2", "D1", "D", "D1"],
        }
    )
    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)
    assert result == {"Episodes": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="389",
    message="Reason episode ceased is that child transferred to care of adult social care services, but child is aged under 16.",
    affected_fields=["REC"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    if "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB16"] = header["DOB"] + pd.DateOffset(years=16)

        episodes_merged = (
            episodes.reset_index()
            .merge(
                header,
                how="left",
                on=["CHILD"],
                suffixes=("", "_header"),
                indicator=True,
            )
            .set_index("index")
        )

        ceased_asc = episodes_merged["REC"].str.upper().astype(str).isin(["E7"])
        ceased_over_16 = episodes_merged["DOB16"] <= episodes_merged["DEC"]

        error_mask = ceased_asc & ~ceased_over_16

        error_locations = episodes_merged.index[error_mask].unique()

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103"],
            "REC": ["E7", "E7", "X1", pd.NA, "E7"],
            "DEC": ["16/03/2021", "17/06/2020", "20/03/2020", pd.NA, "23/08/2020"],
        }
    )

    fake_data_child = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "103"],
            "DOB": ["16/03/2005", "23/09/2002", "31/12/2014", "31/12/2014"],
        }
    )

    fake_dfs = {"Episodes": fake_data, "Header": fake_data_child}

    result = validate(fake_dfs)

    assert result == {"Episodes": [4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="119",
    message="If the decision is made that a child should no longer be placed for adoption, then the date of this decision and the reason why this decision was made must be completed.",
    affected_fields=["REASON_PLACED_CEASED", "DATE_PLACED_CEASED"],
    tables=["PlacedAdoption"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs:
        return {}
    else:
        adopt = dfs["PlacedAdoption"]
        na_placed_ceased = adopt["DATE_PLACED_CEASED"].isna()
        na_reason_ceased = adopt["REASON_PLACED_CEASED"].isna()

        validation_error = (na_placed_ceased & ~na_reason_ceased) | (
            ~na_placed_ceased & na_reason_ceased
        )
        validation_error_locations = adopt.index[validation_error]

        return {"PlacedAdoption": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DATE_PLACED_CEASED": ["22/11/2015", "08/05/2010", pd.NA, pd.NA],
            "REASON_PLACED_CEASED": ["XXX", pd.NA, "10/05/2009", pd.NA],
        }
    )

    fake_dfs = {"PlacedAdoption": fake_data}

    result = validate(fake_dfs)

    assert result == {"PlacedAdoption": [1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


import pandas as pd


@rule_definition(
    code="436",
    message="Reason for new episode is that both child’s placement and legal status have changed, but this is not reflected in the episode data.",
    affected_fields=["RNE", "LS", "PLACE", "PL_POST", "URN", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")

        epi.sort_values(["CHILD", "DECOM"], inplace=True)
        epi.reset_index(inplace=True)
        epi.reset_index(inplace=True)
        epi["LAG_INDEX"] = epi["level_0"].shift(-1)
        epi.fillna(
            value={
                "LS": "*",
                "PLACE": "*",
                "PL_POST": "*",
                "URN": "*",
                "PLACE_PROVIDER": "*",
            },
            inplace=True,
        )
        epi_merge = epi.merge(
            epi,
            how="inner",
            left_on="level_0",
            right_on="LAG_INDEX",
            suffixes=["", "_PRE"],
        )

        epi_multi_row = epi_merge[epi_merge["CHILD"] == epi_merge["CHILD_PRE"]]
        epi_has_B_U = epi_multi_row[epi_multi_row["RNE"].isin(["U", "B"])]

        mask_ls = epi_has_B_U["LS"] == epi_has_B_U["LS_PRE"]

        mask1 = epi_has_B_U["PLACE"] == epi_has_B_U["PLACE_PRE"]
        mask2 = epi_has_B_U["PL_POST"] == epi_has_B_U["PL_POST_PRE"]
        mask3 = epi_has_B_U["URN"] == epi_has_B_U["URN_PRE"]
        mask4 = epi_has_B_U["PLACE_PROVIDER"] == epi_has_B_U["PLACE_PROVIDER_PRE"]

        error_mask = mask_ls | (mask1 & mask2 & mask3 & mask4)

        error_list = epi_has_B_U[error_mask]["index"].to_list()
        error_list.sort()
        return {"Episodes": error_list}


def test_validate():
    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "11",
                "DECOM": "03/06/2020",
                "RNE": "P",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "R",
                "PLACE_PROVIDER": "b",
            },  # 0
            {
                "CHILD": "11",
                "DECOM": "05/06/2020",
                "RNE": "U",
                "LS": "X",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 1
            {
                "CHILD": "11",
                "DECOM": "06/06/2020",
                "RNE": "B",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 2 Fails diff LS, same PLACE et. al
            {
                "CHILD": "11",
                "DECOM": "08/06/2020",
                "RNE": "U",
                "LS": "X",
                "PLACE": "X1",
                "PL_POST": "TT",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 3
            {
                "CHILD": "22",
                "DECOM": "05/06/2020",
                "RNE": "U",
                "LS": "P",
                "PLACE": pd.NA,
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 4
            {
                "CHILD": "33",
                "DECOM": "06/06/2020",
                "RNE": "P",
                "LS": "P",
                "PLACE": "E11",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 5
            {
                "CHILD": "33",
                "DECOM": "07/06/2020",
                "RNE": "B",
                "LS": pd.NA,
                "PLACE": pd.NA,
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 6
            {
                "CHILD": "44",
                "DECOM": "08/06/2020",
                "RNE": "T",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 7
            {
                "CHILD": "44",
                "DECOM": "09/06/2020",
                "RNE": "B",
                "LS": "P",
                "PLACE": "E11",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 8 Fails same LS
            {
                "CHILD": "44",
                "DECOM": "15/06/2020",
                "RNE": "B",
                "LS": "P",
                "PLACE": pd.NA,
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 9 Fails same LS
            {
                "CHILD": "55",
                "DECOM": "11/06/2020",
                "RNE": "P",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 10
            {
                "CHILD": "66",
                "DECOM": "12/06/2020",
                "RNE": "P",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 11
            {
                "CHILD": "66",
                "DECOM": "13/06/2020",
                "RNE": "P",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 12
            {
                "CHILD": "66",
                "DECOM": "14/06/2020",
                "RNE": "B",
                "LS": "X",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 13 Fails diff LS, same PLACE et. al
            {
                "CHILD": "66",
                "DECOM": "15/06/2020",
                "RNE": "P",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 14
            {
                "CHILD": "77",
                "DECOM": "16/06/2020",
                "RNE": "P",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 15
            {
                "CHILD": "77",
                "DECOM": "17/06/2020",
                "RNE": "B",
                "LS": "P",
                "PLACE": "X1",
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 16  Fails same PLACE et. al
            {
                "CHILD": "77",
                "DECOM": "18/06/2020",
                "RNE": "P",
                "LS": pd.NA,
                "PLACE": pd.NA,
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 17
            {
                "CHILD": "99",
                "DECOM": "31/06/2020",
                "RNE": "P",
                "LS": pd.NA,
                "PLACE": pd.NA,
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 18
            {
                "CHILD": "12",
                "DECOM": pd.NA,
                "RNE": "U",
                "LS": pd.NA,
                "PLACE": pd.NA,
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 19
            {
                "CHILD": pd.NA,
                "DECOM": pd.NA,
                "RNE": "U",
                "LS": pd.NA,
                "PLACE": pd.NA,
                "PL_POST": "X1",
                "URN": "a",
                "PLACE_PROVIDER": "b",
            },  # 20
        ]
    )

    fake_dfs = {"Episodes": fake_data}
    result = validate(fake_dfs)
    assert result == {"Episodes": [2, 8, 9, 13, 16]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="196",
    message="Strengths and Difficulties (SDQ) reason is not a valid code.",
    affected_fields=["SDQ_REASON"],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}

    oc2 = dfs["OC2"]
    code_list = ["SDQ1", "SDQ2", "SDQ3", "SDQ4", "SDQ5"]

    mask = oc2["SDQ_REASON"].isin(code_list) | oc2["SDQ_REASON"].isna()

    validation_error_mask = ~mask
    validation_error_locations = oc2.index[validation_error_mask]

    return {"OC2": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SDQ_REASON": ["SDQ2", "sdq2", "", pd.NA],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="577",
    message="Child ceased to be looked after but there is a missing/away from placement without authorisation period without an end date.",
    affected_fields=["MIS_END"],
    tables=["Episodes", "Missing"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Missing" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        missing = dfs["Missing"]

        episodes["original_index"] = episodes.index

        # put dates in appropriate format.
        missing["MIS_END"] = pd.to_datetime(
            missing["MIS_END"], format="%d/%m/%Y", errors="coerce"
        )
        missing["MIS_START"] = pd.to_datetime(
            missing["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        # filter data based on provided conditions.
        missing = missing[missing["MIS_START"].notna()].copy()
        episodes = episodes.dropna(subset=["DECOM"])

        # create period of care blocks
        episodes = episodes.sort_values(["CHILD", "DECOM"])

        episodes["index"] = pd.RangeIndex(0, len(episodes))
        episodes["index+1"] = episodes["index"] + 1
        episodes = episodes.merge(
            episodes,
            left_on="index",
            right_on="index+1",
            how="left",
            suffixes=[None, "_prev"],
        )
        episodes = episodes[
            [
                "original_index",
                "DECOM",
                "DEC",
                "DEC_prev",
                "CHILD",
                "CHILD_prev",
                "LS",
            ]
        ]

        episodes["new_period"] = (episodes["DECOM"] > episodes["DEC_prev"]) | (
            episodes["CHILD"] != episodes["CHILD_prev"]
        )
        episodes["period_id"] = episodes["new_period"].astype(int).cumsum()

        # allocate the same DECOM (min) and DEC (max) to all episodes in a period of care.
        pocs = pd.DataFrame()
        pocs[["CHILD", "poc_DECOM"]] = episodes.groupby("period_id")[
            ["CHILD", "DECOM"]
        ].first()
        pocs["poc_DEC"] = episodes.groupby("period_id")["DEC"].nth(-1)

        # prepare to merge
        missing["index_ing"] = missing.index

        pocs = pocs.merge(missing, on="CHILD", how="right", suffixes=["_eps", "_ing"])
        # poc : period of care
        pocs["out_of_poc"] = (
            # (a) POC is over, but Missing ep is ongoing:
            (
                pocs["MIS_START"].notna()
                & pocs["MIS_END"].isna()
                & pocs["poc_DEC"].notna()
            )
            # (b) Ongoing Missing ep, but started before POC:
            | (pocs["MIS_END"].isna() & (pocs["MIS_START"] < pocs["poc_DECOM"]))
            # (c) Missing ep ends after POC:
            | (pocs["MIS_END"] > pocs["poc_DEC"])
            # (d) Missing ep ends before POC:
            | (pocs["MIS_END"] < pocs["poc_DECOM"])
            # (e?) Starts during a previous poc??? (row 12 of Missing table in testvalidate_577)
        )

        # Drop rows where child was not present in 'Missing' table.
        pocs = pocs[pocs["index_ing"].notna()]

        mask = pocs.groupby("index_ing")["out_of_poc"].transform("min")
        miss_error_locs = pocs.loc[mask, "index_ing"].astype(int).unique().tolist()

        return {"Missing": miss_error_locs}


def test_validate():
    import pandas as pd

    fake_data_mis = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "MIS_START": "01/01/1981",
                "MIS_END": pd.NA,
            },  # 0
            {
                "CHILD": "2",
                "MIS_START": "01/01/1981",
                "MIS_END": "02/06/2020",
            },  # 1
            {
                "CHILD": "3",
                "MIS_START": "01/01/1981",
                "MIS_END": "03/06/2022",
            },  # 2 fail: fail1 fail2
            {
                "CHILD": "3",
                "MIS_START": "01/01/1981",
                "MIS_END": "03/06/2020",
            },  # 3 pass: fail with the first, pass with the second
            {
                "CHILD": "4",
                "MIS_START": "01/01/1981",
                "MIS_END": "04/06/2020",
            },  # 4
            {
                "CHILD": "5",
                "MIS_START": "01/01/1981",
                "MIS_END": "01/01/1982",
            },  # 5 passes with full period of care DECOM-DEC (7/8) and 9
            {
                "CHILD": "5",
                "MIS_START": "01/01/1981",
                "MIS_END": "01/01/2000",
            },  # 6 pass: fails with first set of episodes(7/8 poc), but passes with next period of care (9)
            {
                "CHILD": "6",
                "MIS_START": "01/01/1981",
                "MIS_END": "01/01/1981",
            },  # 7 fail when compared with DECOM
            {
                "CHILD": "7",
                "MIS_START": pd.NA,
                "MIS_END": "01/01/1981",
            },  # 8 ignore fail: MIS_START is NaN
            {
                "CHILD": "8",
                "MIS_START": "01/01/1981",
                "MIS_END": pd.NA,
            },  # 9 fail: MIS_END is nan
            {
                "CHILD": "9",
                "MIS_START": "01/01/2020",
                "MIS_END": pd.NA,
            },  # 10 fail: MIS_END is nan
            {
                "CHILD": "9",
                "MIS_START": "01/01/2021",
                "MIS_END": pd.NA,
            },  # 11 pass: ongoing mis-ep during ongoing poc
            {
                "CHILD": "9",
                "MIS_START": "04/06/2020",
                "MIS_END": "04/06/2021",
            },  # 12 pass: ends during a poc (?!)
        ]
    )
    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "DECOM": "01/04/1977",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
            },  # 0
            {
                "CHILD": "2",
                "DECOM": "01/04/1979",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": pd.NA,
            },  # 1
            {
                "CHILD": "88",
                "DECOM": "01/05/1980",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "xx",
            },  # 2
            {
                "CHILD": "3",
                "DECOM": "01/04/1980",
                "DEC": "01/01/1981",
                "LS": "V3",
                "REC": "kk",
            },  # 3
            {
                "CHILD": "3",
                "DECOM": "01/04/2020",
                "DEC": "01/01/2021",
                "LS": "V3",
                "REC": "kk",
            },  # 4
            {
                "CHILD": "3",
                "DECOM": "01/01/2021",
                "DEC": "05/10/2021",
                "LS": "V3",
                "REC": "kk",
            },  # 5
            {
                "CHILD": "4",
                "DECOM": "01/04/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
            },  # 6
            {
                "CHILD": "5",
                "DECOM": "01/04/1978",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "xx",
            },  # 7 --poc start
            {
                "CHILD": "5",
                "DECOM": "01/01/1981",
                "DEC": "01/01/1990",
                "LS": "o",
                "REC": "xx",
            },  # 8  poc end--
            {
                "CHILD": "5",
                "DECOM": "01/04/1983",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "kk",
            },  # 9
            {
                "CHILD": "6",
                "DECOM": "01/04/1983",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "kk",
            },  # 10
            {
                "CHILD": "7",
                "DECOM": "01/04/1983",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "kk",
            },  # 11
            {
                "CHILD": "8",
                "DECOM": "01/04/1983",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "xx",
            },  # 12
            {
                "CHILD": "9",
                "DECOM": "01/01/2020",
                "DEC": "04/06/2020",
                "LS": "o",
                "REC": "<--",
            },  # 10
            {
                "CHILD": "9",
                "DECOM": "04/06/2020",
                "DEC": "11/08/2020",
                "LS": "o",
                "REC": "-",
            },  # 10
            {
                "CHILD": "9",
                "DECOM": "11/08/2020",
                "DEC": "01/12/2020",
                "LS": "o",
                "REC": "-->",
            },  # 10
            {
                "CHILD": "9",
                "DECOM": "22/12/2020",
                "DEC": "25/12/2020",
                "LS": "o",
                "REC": "<->",
            },  # 10
            {
                "CHILD": "9",
                "DECOM": "01/01/2021",
                "DEC": "22/02/2021",
                "LS": "o",
                "REC": "<--",
            },  # 10
            {
                "CHILD": "9",
                "DECOM": "22/02/2021",
                "DEC": "08/08/2021",
                "LS": "V3",
                "REC": "-",
            },  # 10
            {
                "CHILD": "9",
                "DECOM": "08/08/2021",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "--o",
            },  # 10
        ]
    )

    fake_dfs = {"Episodes": fake_data_eps, "Missing": fake_data_mis}

    result = validate(fake_dfs)

    assert result == {"Missing": [0, 1, 2, 7, 9, 10]}



import logging
from io import BytesIO

import pandas as pd
from openpyxl import Workbook
from openpyxl.utils import get_column_letter
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.worksheet.table import Table, TableStyleInfo

logger = logging.getLogger(__name__)


class Report:
    def __init__(self, data_store, ruleset_registry):
        self.__child_report = None
        self.__child_summary = None
        self.__error_report = None
        self.__error_summary = None

        self._errors_df = pd.DataFrame(
            [
                (rule.code, rule.message, rule.affected_fields)
                for _, rule in ruleset_registry.items()
            ],
            columns=[
                "Code",
                "Description",
                "Fields",
            ],
        ).sort_values("Code")
        logger.debug(
            f"Created error summary df with {self._errors_df.shape[0]} configured errors."
        )

        dataframes = []
        for table, value in data_store.items():
            if table == "metadata":
                # report should only contain child-level data.
                continue
            cols_error = [c for c in value.columns if c[:4] == "ERR_"]
            cols_data = [
                c
                for c in value.columns
                if c[:4] != "ERR_" and c[:2] != "__" and c != "CHILD"
            ]

            df_error = value[["CHILD"] + cols_error].copy()
            df_error["Table"] = table
            df_error["RowID"] = df_error.index
            df_error["Context"] = value[cols_data].apply(lambda row: {**row}, axis=1)

            dataframes.append(df_error)

        report = pd.concat(dataframes, ignore_index=True)
        self.child_cols = ["Table", "RowID", "CHILD", "Context"]
        self.error_cols = [c for c in report.columns if c not in self.child_cols]
        report["Total Errors"] = report[self.error_cols].sum(axis=1)
        self.report = report[self.child_cols + self.error_cols + ["Total Errors"]]

    @property
    def error_report(self):
        """
        The error report is a simple report containing the number of errors report for each error code
        as well as the error description and the affected fields.
        """
        if self.__error_report is None:
            df = pd.DataFrame(
                self.report[self.error_cols].sum(), columns=["Count"]
            ).reset_index()
            df["Code"] = df["index"].str[4:]
            df = df[["Code", "Count"]]
            self.__error_report = df.merge(
                self._errors_df, on="Code", how="left"
            ).sort_values("Code")[
                [
                    "Code",
                    "Description",
                    "Fields",
                    "Count",
                ]
            ]
        return self.__error_report

    @property
    def error_summary(self):
        if self.__error_summary is None:
            headers = ["Code", "Error Count", "Error Description", "Affected Fields"]
            df = self.error_report.rename(
                columns={
                    "Description": "Error Description",
                    "Count": "Error Count",
                }
            )
            df["Affected Fields"] = df["Fields"].apply(lambda x: ", ".join(x))
            self.__error_summary = df[headers]

        return self.__error_summary

    @property
    def child_report(self):
        """
        The child report contains details of every error reported, including context information with the data
        from the affected row.
        """
        if self.__child_report is None:
            child_report = self.report.melt(
                id_vars=self.child_cols,
                value_vars=self.error_cols,
                var_name="ErrorColumn",
                value_name="IsError",
            )
            child_report = child_report[child_report.IsError == True].reset_index()
            child_report["Code"] = child_report.ErrorColumn.str[4:]

            child_error_count = (
                child_report[["CHILD", "Code"]]
                .groupby("CHILD")
                .agg(**{"Child Error Count": ("Code", "count")})
                .reset_index()
            )

            child_report = child_report.merge(child_error_count, on="CHILD", how="left")

            error_report = self.error_report.rename(
                columns=dict(Count="Error Type Count")
            )
            child_report = child_report.merge(error_report, on=["Code"], how="left")
            child_report = child_report[
                [
                    "CHILD",
                    "Table",
                    "RowID",
                    "Context",
                    "Code",
                    "Description",
                    "Fields",
                    "Child Error Count",
                    "Error Type Count",
                ]
            ]
            self.__child_report = child_report
        return self.__child_report

    @property
    def child_summary(self):
        if self.__child_summary is None:
            self.__child_summary = _create_child_summary(self.child_report)
        return self.__child_summary

    def excel_report(self, file=None):
        """
        Creates an excel report containing an error summary and full details of each error with context information.

        If a string or file-like is provided, then the output is saved to this,
        otherwise the workbook data are returned as a buffer.
        """
        logger.info("Creating workbook")
        wb = Workbook(write_only=True)

        error_sheet = wb.create_sheet(title="Error Summary")
        _populate_error_sheet(error_sheet, self.error_summary)

        child_sheet = wb.create_sheet(title="Child Summary")
        _populate_child_sheet(child_sheet, self.child_summary)

        logger.info("Creating workbook - saving")
        if file:
            wb.save(file)
        else:
            io = BytesIO()
            wb.save(io)
            return io.getbuffer()

    def csv_report(self, sheetname, file=None):
        if sheetname == "errors":
            df = self.error_summary
        elif sheetname == "children":
            df = self.child_summary
        else:
            raise ValueError("Unknown export type requested")

        if file:
            df.to_csv(file, index=False)
        else:
            return df.to_csv(index=False)


def _populate_error_sheet(error_sheet, df):
    logger.debug("Creating workbook - error summary")

    error_sheet.column_dimensions["A"].width = 15
    error_sheet.column_dimensions["B"].width = 15
    error_sheet.column_dimensions["C"].width = 60
    error_sheet.column_dimensions["D"].width = 120

    headers = [c for c in df.columns]
    error_sheet.append(headers)
    for r in dataframe_to_rows(df, index=False, header=False):
        error_sheet.append(r)

    error_table = Table(
        displayName="Errors",
        ref=f"A1:{get_column_letter(len(headers))}{df.shape[0] + 1}",
    )
    error_table._initialise_columns()
    for column, value in zip(error_table.tableColumns, headers):
        column.name = value

    # Add a default style with striped rows and banded columns
    style = TableStyleInfo(name="TableStyleMedium9", showRowStripes=True)
    error_table.tableStyleInfo = style
    error_sheet.add_table(error_table)


def _populate_child_sheet(child_sheet, df):
    logger.debug("Creating workbook - child summary")

    child_sheet.column_dimensions["A"].width = 15
    child_sheet.column_dimensions["B"].width = 15
    child_sheet.column_dimensions["C"].width = 15
    child_sheet.column_dimensions["D"].width = 15
    child_sheet.column_dimensions["E"].width = 60
    child_sheet.column_dimensions["F"].width = 30
    child_sheet.column_dimensions["G"].width = 40
    child_sheet.column_dimensions["H"].width = 20
    child_sheet.column_dimensions["I"].width = 15
    child_sheet.column_dimensions["J"].width = 15

    logger.debug("Child Summary - Adding headers")
    headers = [c for c in df.columns]
    child_sheet.append(headers)

    logger.debug("Child Summary - Populating rows")
    for index, row in df.iterrows():
        child_sheet.append([row[h] for h in headers])

    logger.debug("Child Summary - Adding table")
    child_table = Table(
        displayName="ChildSummary",
        ref=f"A1:{get_column_letter(len(headers))}{df.shape[0] + 1}",
    )
    child_table._initialise_columns()
    for column, value in zip(child_table.tableColumns, headers):
        column.name = value

    style = TableStyleInfo(name="TableStyleMedium9", showRowStripes=True)
    child_table.tableStyleInfo = style
    child_sheet.add_table(child_table)


def _create_child_summary(child_report: pd.DataFrame) -> pd.DataFrame:
    logger.debug("Child Summary - Renaming Columns")
    headers = [
        "Child",
        "Affected Table",
        "Error Row",
        "Error Code",
        "Error Description",
        "Error Fields",
        "Affected Values",
        "Locator Hints",
        "Errors for Child",
        "Errors of Type",
    ]
    df = child_report.rename(
        columns={
            "CHILD": "Child",
            "Table": "Affected Table",
            "RowID": "Error Row",
            "Code": "Error Code",
            "Description": "Error Description",
            "Child Error Count": "Errors for Child",
            "Error Type Count": "Errors of Type",
        }
    )
    logger.debug("Child Summary - Setting Error Fields")
    df["Error Fields"] = df["Fields"].apply(lambda x: ", ".join(x))
    logger.debug("Child Summary - Setting Affected Values")
    df["Affected Values"] = df[["Fields", "Context"]].apply(
        lambda row: ", ".join(
            [f"{f}: {row.Context[f]}" for f in row.Fields if f in row.Context]
        ),
        axis=1,
    )
    logger.debug("Child Summary - Setting Locator Hints")
    df["Locator Hints"] = df.Context.apply(
        lambda row: ", ".join(
            [f"{f}: {row[f]}" for f in ["DECOM", "MIS_START", "REVIEW"] if f in row]
        )
    )
    logger.debug("Child Summary - Sorting")
    df.sort_values(["Child", "Affected Table", "Error Code"], inplace=True)
    return df[headers]



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="182",
    message="Data entries on immunisations, teeth checks, health assessments and substance misuse problem identified should be completed or all OC2 fields should be left blank.",
    affected_fields=[
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
        "CONVICTED",
        "HEALTH_CHECK",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]

        mask1 = (
            oc2["IMMUNISATIONS"].isna()
            | oc2["TEETH_CHECK"].isna()
            | oc2["HEALTH_ASSESSMENT"].isna()
            | oc2["SUBSTANCE_MISUSE"].isna()
        )
        mask2 = (
            oc2["CONVICTED"].isna()
            & oc2["HEALTH_CHECK"].isna()
            & oc2["INTERVENTION_RECEIVED"].isna()
            & oc2["INTERVENTION_OFFERED"].isna()
        )

        validation_error = mask1 & ~mask2
        validation_error_locations = oc2.index[validation_error]

        return {"OC2": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "IMMUNISATIONS": [pd.NA, pd.NA, pd.NA, "1", pd.NA, "1", pd.NA, "1"],
            "TEETH_CHECK": [pd.NA, pd.NA, pd.NA, "1", pd.NA, "1", "1", "1"],
            "HEALTH_ASSESSMENT": [pd.NA, pd.NA, pd.NA, "1", pd.NA, "1", pd.NA, "1"],
            "SUBSTANCE_MISUSE": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "1", "1", "1"],
            "CONVICTED": [pd.NA, "1", pd.NA, pd.NA, pd.NA, "1", "1", pd.NA],
            "HEALTH_CHECK": [pd.NA, pd.NA, "1", pd.NA, pd.NA, "1", "1", pd.NA],
            "INTERVENTION_RECEIVED": [pd.NA, pd.NA, pd.NA, "1", pd.NA, "1", "1", pd.NA],
            "INTERVENTION_OFFERED": [pd.NA, pd.NA, pd.NA, pd.NA, "1", "1", "1", pd.NA],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)
    assert result == {"OC2": [1, 2, 3, 4, 6]}



from lac_validator.rule_engine import rule_definition


import pandas as pd


@rule_definition(
    code="1002",
    message="This child has no previous episodes of care, therefore should not have care leaver information recorded. "
    "[NOTE: This tool can only test the current and previous year data loaded into the tool - this "
    "check may generate false positives if a child had episodes prior to last year's collection.]",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["Episodes", "OC3", "Episodes_last"],
)
def validate(dfs):
    if "Episodes" not in dfs or "OC3" not in dfs or "Episodes_last" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        oc3 = dfs["OC3"]

        episodes_last = dfs["Episodes_last"]
        has_previous_episodes = oc3["CHILD"].isin(episodes_last["CHILD"])
        has_current_episodes = oc3["CHILD"].isin(episodes["CHILD"])
        error_mask = ~has_current_episodes & ~has_previous_episodes

        validation_error_locations = oc3.index[error_mask]

        return {"OC3": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_episodes_prev = pd.DataFrame(
        {
            "CHILD": ["101", "101", "102", "102", "103"],
            "DEC": ["20/06/2020", pd.NA, pd.NA, "14/03/2021", "06/08/2020"],
        }
    )

    fake_episodes = pd.DataFrame({"CHILD": ["101"], "DEC": ["20/05/2021"]})

    fake_oc3 = pd.DataFrame({"CHILD": ["101", "102", "103", "104"]})

    fake_dfs = {
        "Episodes": fake_episodes,
        "Episodes_last": fake_episodes_prev,
        "OC3": fake_oc3,
    }
    result = validate(fake_dfs)
    assert result == {"OC3": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="398",
    message="Distance field completed but child looked after under legal status V3 or V4.",
    affected_fields=["LS", "HOME_POST", "PL_POST"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        mask = ((episodes["LS"] == "V3") | (episodes["LS"] == "V4")) & (
            episodes["HOME_POST"].notna() | episodes["PL_POST"].notna()
        )
        error_locations = episodes.index[mask]
        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "HOME_POST": [
                "AB1 0JD",
                "invalid",
                "AB1 0JD",
                "invalid",
                "AB10JD",
            ],
            "LS": ["V3", "U1", "V4", "T1", "U1"],
            "PL_POST": [
                "AB1 0JD",
                "AB1 0JD",
                pd.NA,
                "invalid",
                "AB1 0JD",
            ],
        }
    )
    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="624",
    message="Date of birth of the first child contradicts the date of birth of the first child previously "
    + "recorded.",
    affected_fields=["MC_DOB"],
    tables=["Header", "Header_last"],
)
def validate(dfs):
    if "Header" not in dfs or "Header_last" not in dfs:
        return {}
    else:
        hea = dfs["Header"]
        hea_pre = dfs["Header_last"]
        hea["orig_idx"] = hea.index

        err_co = hea.merge(
            hea_pre, how="inner", on="CHILD", suffixes=["", "_PRE"]
        ).query("MC_DOB_PRE.notna() & (MC_DOB != MC_DOB_PRE)")

        err_list = err_co["orig_idx"].unique().tolist()
        err_list.sort()
        return {"Header": err_list}


def test_validate():
    import pandas as pd

    hdr = pd.DataFrame(
        [
            {"CHILD": "111", "MC_DOB": "01/06/2020"},  # 0
            {"CHILD": "222", "MC_DOB": "04/06/2020"},  # 1
            {"CHILD": "333", "MC_DOB": pd.NA},  # 2
            {"CHILD": "444", "MC_DOB": "08/09/2020"},  # 3
            {"CHILD": "555", "MC_DOB": pd.NA},  # 4
            {"CHILD": "66", "MC_DOB": "01/02/2020"},  # 5
        ]
    )
    hdr_last = pd.DataFrame(
        [
            {"CHILD": "111", "MC_DOB": "01/06/2020"},  # 0
            {"CHILD": "222", "MC_DOB": "04/06/2020"},  # 1
            {"CHILD": "333", "MC_DOB": "01/06/2019"},  # 2
            {"CHILD": "444", "MC_DOB": "10/09/2020"},  # 3
            {"CHILD": "555", "MC_DOB": pd.NA},  # 4
            {"CHILD": "66", "MC_DOB": pd.NA},  # 5
        ]
    )
    fake_dfs = {"Header": hdr, "Header_last": hdr_last}

    result = validate(fake_dfs)

    assert result == {"Header": [2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="133",
    message="Data entry for accommodation after leaving care is invalid. If reporting on a childs accommodation after leaving care the data entry must be valid",
    affected_fields=["ACCOM"],
    tables=["OC3"],
)
def validate(dfs):
    if "OC3" not in dfs:
        return {}

    else:
        oc3 = dfs["OC3"]
        valid_codes = [
            "B1",
            "B2",
            "C1",
            "C2",
            "D1",
            "D2",
            "E1",
            "E2",
            "G1",
            "G2",
            "H1",
            "H2",
            "K1",
            "K2",
            "R1",
            "R2",
            "S2",
            "T1",
            "T2",
            "U1",
            "U2",
            "V1",
            "V2",
            "W1",
            "W2",
            "X2",
            "Y1",
            "Y2",
            "Z1",
            "Z2",
            "0",
        ]

        error_mask = ~oc3["ACCOM"].isin(valid_codes)

        error_locations = oc3.index[error_mask]

        return {"OC3": error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "ACCOM": ["B1", "B2", pd.NA, "s2", "K1", "X", "1"],
        }
    )

    fake_dfs = {"OC3": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC3": [2, 3, 5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="520",
    message="Data entry on the legal status of adopters shows different gender married couple but data entry on genders of adopters shows it as a same gender couple.",
    affected_fields=["LS_ADOPTR", "SEX_ADOPTR"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}
    else:
        ad1 = dfs["AD1"]
        # check condition
        mask = (ad1["LS_ADOPTR"] == "L11") & (ad1["SEX_ADOPTR"] != "MF")
        error_locations = ad1.index[mask]
        return {"AD1": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_ad1 = pd.DataFrame(
        {
            "LS_ADOPTR": ["L11", "L11", pd.NA, "L4", "L11", "L1", "L4"],
            "SEX_ADOPTR": ["MM", "FF", "MM", "M1", "MF", "F1", "xxxxx"],
        }
    )
    fake_dfs = {"AD1": fake_ad1}

    result = validate(fake_dfs)
    assert result == {"AD1": [0, 1]}



import importlib
import os
from pathlib import Path
import pandas as pd


import click
import pytest

from lac_validator import lac_validator
from lac_validator.ingress import read_from_text
from lac_validator.report import Report
from lac_validator.utils import process_uploaded_files
from lac_validator.config import column_names


@click.group()
def cli():
    pass


# LIST
@cli.command(name="list")
@click.option(
    "--ruleset",
    "-r",
    default="lac2023_24",
    help="validation year, e.g lac2023_24",
)
def list_cmd(ruleset):
    """
    :param str ruleset: validation year whose version of rules should be run.

    :return cli output: list of rules in validation year.
    """
    module = importlib.import_module(f"lac_validator.rules.{ruleset}")
    ruleset_registry = getattr(module, "registry")
    for _, rule in ruleset_registry.items():
        click.echo(f"{rule.code}\t{rule.message}")


# Output rules, message, affected fields to csv
@cli.command(name="rule-info")
@click.option(
    "--ruleset",
    "-r",
    default="lac2024_25",
    help="validation year, e.g lac2024_25",
)
def rule_info_cmd(ruleset):
    """
    :param str ruleset: validation year whose version of rules should be run.

    :return cli output: tables of rules to be run in a year with message, affected fields,
                        affected tables, and CSV download of info.
    """
    module = importlib.import_module(f"lac_validator.rules.{ruleset}")
    ruleset_registry = getattr(module, "registry")
    rules_list = []
    for _, rule in ruleset_registry.items():
        rules_list.append(
            {
                "Rule code": f"{rule.code}",
                "Rule message": f"{rule.message}",
                "Affected fields": f"{sorted(rule.affected_fields)}",
                "Tables": f"{sorted(rule.tables)}",
            }
        )
    rule_info = pd.DataFrame(rules_list)
    rule_info.sort_values("Rule code", inplace=True)
    click.echo(rule_info)

    # uncomment to check if any rules are midding table info
    # click.echo(rule_info[rule_info["Tables"].isna()])

    rule_info.to_csv(f"output_data/903_{ruleset[3:]}_rule_info.csv", index=False)


# TEST
@cli.command(name="test")
@click.option(
    "--ruleset",
    "-r",
    default="lac2023_24",
    help="validation year, e.g lac2023_24",
)
def test_cmd(ruleset):
    """
    Runs pytest of rules specified
    :param str ruleset: validation year whose rules should be run
    :return: classic pytest output
    """
    module = importlib.import_module(f"lac_validator.rules.{ruleset}")
    module_folder = Path(module.__file__).parent
    # May 2023. There are 288 rule files.
    test_files = [
        str(p.absolute()) for p in module_folder.glob("*.py") if p.stem != "__init__"
    ]
    pytest.main(test_files)


# TEST one rule
@cli.command(name="test_one_rule")
@click.argument("code", type=str, required=True)
@click.option(
    "--ruleset",
    "-r",
    default="lac2023_24",
    help="validation year, e.g lac2023_24",
)
def test_one_rule(code, ruleset):
    """
    Runs pytest of rules specified
    :param str ruleset: validation year whose rules should be run
    :return: classic pytest output
    """
    module = importlib.import_module(f"lac_validator.rules.{ruleset}")
    module_folder = Path(module.__file__).parent

    file_path = os.path.join(module_folder, f"rule_{code}.py")

    pytest.main([file_path])


# RUN
@cli.command(name="run")
@click.argument("p4a_path", type=click.File("rt"), required=True)
@click.argument("ad1_path", type=click.File("rt"), required=True)
@click.option(
    "--ruleset",
    "-r",
    default="lac2023_24",
    help="validation year e.g lac2022_23",
)
@click.option("--select", "-s", default=None)
def run_all(p4a_path, ad1_path, ruleset, select):
    """
    created with code from offlinedebug.py

    CLI command:
    python -m lac_validator run <filepath_>

    :param str ruleset: validation year.
    :param str select: code of specific rule that should be run.
    """
    # p4a_path = "tests\\fake_data\placed_for_adoption_errors.csv"
    # ad1_path = "tests\\fake_data\\ad1.csv"
    # frontend_files_dict = {"This year":[p4a_path, ad1_path], "Prev year": [p4a_path]}

    frontend_files_dict = {"This year": [p4a_path, ad1_path]}
    files_list = process_uploaded_files(frontend_files_dict)

    # the rest of the metadata is added in read_from_text() when instantiating Validator
    metadata = {"collectionYear": "2022", "localAuthority": "E09000027"}
    module = importlib.import_module(f"lac_validator.rules.{ruleset}")
    ruleset_registry = getattr(module, "registry")

    v = lac_validator.LacValidator(
        metadata=metadata,
        files=files_list,
        registry=ruleset_registry,
        selected_rules=None,
    )
    results = v.ds_results

    click.echo(v.ds_results)
    click.echo(f"skipped {v.skips}")
    click.echo(f"done: {v.dones}")

    r = Report(results, ruleset_registry)
    # click.echo(f"*****************Error report******************")
    # click.echo(r.error_report)
    # click.echo(f"****************Error summary******************")
    # click.echo(r.error_summary)
    # full_issue_df = lac_validator.create_issue_df(r.report, r.error_report)
    # click.echo(f"*****************full issue df******************")
    # click.echo(full_issue_df)


# RUN
@cli.command(name="run-offline")
@click.argument("filename", required=True)
@click.option(
    "--ruleset",
    "-r",
    default="lac2023_24",
    help="validation year e.g lac2023_24",
)
@click.option("--select", "-s", default=None)
def run_all(filename: str, ruleset, select):
    """
    CLI command to run the validator offline, primarily to test ingress

    CLI command:
    python -m lac_validator run-offline <filepath_>

    :param str ruleset: validation year.
    :param str select: code of specific rule that should be run.
    """
    ad1 = f"{filename}/ad1.csv"
    episodes = f"{filename}/episodes.csv"
    header = f"{filename}/header.csv"
    missing = f"{filename}/missing.csv"
    oc2 = f"{filename}/oc2.csv"
    oc3 = f"{filename}/oc3.csv"
    p4a = f"{filename}/placed_for_adoption_errors.csv"
    previous_permanence = f"{filename}/previous_permanence.csv"
    reviews = f"{filename}/reviews.csv"
    uasc = f"{filename}/uasc.csv"
    sw_episodes = f"{filename}/sw_episodes.csv"

    frontend_files_dict = {
        "This year": [
            ad1,
            episodes,
            header,
            missing,
            oc2,
            oc3,
            p4a,
            previous_permanence,
            reviews,
            uasc,
            sw_episodes,
        ]
    }
    files_list = process_uploaded_files(frontend_files_dict)

    # the rest of the metadata is added in read_from_text() when instantiating Validator
    metadata = {"collectionYear": "2023", "localAuthority": "E09000027"}
    module = importlib.import_module(f"lac_validator.rules.{ruleset}")
    ruleset_registry = getattr(module, "registry")

    v = lac_validator.LacValidator(
        metadata=metadata,
        files=files_list,
        registry=ruleset_registry,
        selected_rules=None,
    )

    click.echo(v.dfs)

    results = v.ds_results

    r = Report(results, ruleset_registry)
    full_issue_df = lac_validator.create_issue_df(r.report, r.error_report)
    click.echo(full_issue_df)


# XML to tables
@cli.command(name="xmltocsv")
@click.argument("p4a_path", type=click.File("rt"), required=True)
def xmltocsv(p4a_path):
    with open(p4a_path.name, "rb") as f:
        p4a_filetext = f.read()
    files_list = [
        dict(name=p4a_path.name, description="This year", file_content=p4a_filetext),
    ]

    data_files, _ = read_from_text(files_list)
    click.echo(data_files)


if __name__ == "__main__":
    cli()



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.rules.rule_utils import decom_before_dob


@rule_definition(
    code="374",
    message="Child in residential employment should be at least 14 years old.",
    affected_fields=["DECOM", "PLACE"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs:
        return {}
    else:
        return decom_before_dob(dfs, p_code="P3", y_gap=14)


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "666", "777"],
            "DECOM": [
                "01/01/2020",
                "01/02/2020",
                "02/03/2020",
                "15/03/2010",
                pd.NA,
                "15/03/2004",
                "14/03/2014",
            ],
            "PLACE": ["P3", "P3", "P3", "P3", "P3", "P3", "P3"],  # day before 14bd
            "DEC": [
                "13/02/2020",
                "14/03/2020",
                "14/04/2020",
                pd.NA,
                "27/04/2010",
                "26/04/2004",
                "25/04/2014",
            ],
        }
    )
    fake_hea = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "666", "777"],
            "DOB": [
                "01/01/1998",
                "01/02/2000",
                "02/03/2015",
                "15/01/2010",
                "15/03/2000",
                "15/03/2000",
                "15/03/2000",
            ],
        }
    )

    fake_hea["DOB"] = pd.to_datetime(
        fake_hea["DOB"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi["DECOM"] = pd.to_datetime(
        fake_epi["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi["DEC"] = pd.to_datetime(
        fake_epi["DEC"], format="%d/%m/%Y", errors="coerce"
    )

    fake_dfs = {"Episodes": fake_epi, "Header": fake_hea}

    assert validate(fake_dfs) == {"Episodes": [2, 5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="612",
    message="Date of birth field has been completed but mother field indicates child is not a mother.",
    affected_fields=["SEX", "MOTHER", "MC_DOB"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        header = dfs["Header"]

        error_mask = (
            ((header["MOTHER"].astype(str) == "0") | header["MOTHER"].isna())
            & (header["SEX"].astype(str) == "2")
            & header["MC_DOB"].notna()
        )

        validation_error_locations = header.index[error_mask]

        return {"Header": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SEX": [2, "2", "2", 2, 2, 1, 1, 1, "2"],
            "MOTHER": ["0", "0", pd.NA, pd.NA, 1, 0, 1, pd.NA, pd.NA],
            "MC_DOB": [
                "19/02/2016",
                "dd/mm/yyyy",
                "31/31/19",
                pd.NA,
                "19/02/2019",
                pd.NA,
                "12/10/2010",
                "21/3rd/yYyYyY",
                "21/3rd/yYyYyY",
            ],
        }
    )

    fake_dfs = {"Header": fake_data}

    result = validate(fake_dfs)

    assert result == {"Header": [0, 1, 2, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="165",
    message="Data entry for mother status is invalid.",
    affected_fields=["MOTHER", "SEX", "ACTIV", "ACCOM", "IN_TOUCH", "DECOM"],
    tables=["Header", "Episodes", "OC3"],
)
def validate(dfs):
    if "Header" not in dfs or "Episodes" not in dfs or "OC3" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]
        oc3 = dfs["OC3"]
        collection_start = dfs["metadata"]["collection_start"]
        collection_end = dfs["metadata"]["collection_end"]
        valid_values = ["0", "1"]

        # prepare to merge
        oc3.reset_index(inplace=True)
        header.reset_index(inplace=True)
        episodes.reset_index(inplace=True)

        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        # Drop all episodes with V3/V4 legal status
        v3v4_ls = episodes["LS"].str.upper().isin(["V3", "V4"])
        index_v3v4_ls = episodes.loc[v3v4_ls].index
        episodes.drop(index_v3v4_ls, inplace=True)

        # fill in missing DECs with the collection year end date
        missing_last_DECs = episodes["DEC"].isna()
        episodes.loc[missing_last_DECs, "DEC"] = collection_end

        episodes["EPS"] = (episodes["DEC"] >= collection_start) & (
            episodes["DECOM"] <= collection_end
        )
        episodes["EPS_COUNT"] = episodes.groupby("CHILD")["EPS"].transform("sum")

        merged = episodes.merge(
            header, on="CHILD", how="left", suffixes=["_eps", "_er"]
        ).merge(oc3, on="CHILD", how="left")

        # Raise error if provided <MOTHER> is not a valid value
        value_validity = merged["MOTHER"].notna() & (
            ~merged["MOTHER"].astype(str).isin(valid_values)
        )

        # Raise error if provided <MOTHER> and child is male
        male = merged["MOTHER"].notna() & (merged["SEX"].astype(str) == "1")

        # Raise error if female and not provided
        female = merged["SEX"].astype(str) == "2"
        has_mother = merged["MOTHER"].notna()
        eps_in_year = merged["EPS_COUNT"] > 0
        has_oc3 = (
            merged["ACTIV"].notna()
            | merged["ACCOM"].notna()
            | merged["IN_TOUCH"].notna()
        )

        # If provided <MOTHER> must be a valid value (and child must be female). If not provided <MOTHER> then either <GENDER> is male or no episode record for current year and any of <IN_TOUCH>, <ACTIV> or <ACCOM> have been provided
        mask = (
            value_validity
            | male
            | (~has_mother & female & eps_in_year)
            | (has_mother & female & ~eps_in_year & has_oc3)
        )

        # That is, if value not provided and child is a female with eps in current year or no values of IN_TOUCH, ACTIV and ACCOM, then raise error.
        error_locs_eps = merged.loc[mask, "index_eps"]
        error_locs_header = merged.loc[mask, "index_er"]
        error_locs_oc3 = merged.loc[mask, "index"]

        return {
            "Header": error_locs_header.dropna().unique().tolist(),
            "OC3": error_locs_oc3.dropna().unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": [101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111],
            "IN_TOUCH": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "YES",
                pd.NA,
                pd.NA,
                pd.NA,
            ],
            "ACTIV": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "XXX",
                pd.NA,
                pd.NA,
                pd.NA,
            ],
            "ACCOM": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "XXX",
                pd.NA,
                pd.NA,
            ],
        }
    )
    fake_data_episodes = pd.DataFrame(
        [
            {"CHILD": 101, "DECOM": "01/01/2020", "DEC": "01/05/2020", "LS": "C2"},  # 0
            {"CHILD": 102, "DECOM": "11/01/2020", "DEC": "11/05/2020", "LS": "C2"},  # 1
            {"CHILD": 103, "DECOM": "30/03/2020", "DEC": "30/05/2020", "LS": "C2"},  # 2
            {"CHILD": 103, "DECOM": "30/03/2020", "DEC": "30/05/2020", "LS": "C2"},  # 2
            {"CHILD": 104, "DECOM": "01/01/2020", "DEC": "01/05/2020", "LS": "C2"},  # 3
            {"CHILD": 105, "DECOM": "11/05/2020", "DEC": "11/08/2020", "LS": "C2"},  # 4
            {"CHILD": 105, "DECOM": "01/01/2020", "DEC": "01/05/2020", "LS": "C2"},  # 5
            {"CHILD": 106, "DECOM": "22/01/2020", "DEC": "22/05/2020", "LS": "C2"},  # 6
            {"CHILD": 107, "DECOM": "11/01/2020", "DEC": "11/05/2020", "LS": "C2"},  # 7
            {"CHILD": 108, "DECOM": "22/01/2020", "DEC": "22/05/2020", "LS": "C2"},  # 8
            {"CHILD": 109, "DECOM": "25/03/2020", "DEC": "25/03/2020", "LS": "C2"},  # 9
            {
                "CHILD": 110,
                "DECOM": "01/01/2020",
                "DEC": "01/05/2020",
                "LS": "V3",
            },  # 10
            {
                "CHILD": 110,
                "DECOM": "01/11/2021",
                "DEC": "01/11/2021",
                "LS": "C2",
            },  # 11
            {
                "CHILD": 111,
                "DECOM": "01/11/2020",
                "DEC": "01/11/2020",
                "LS": "V3",
            },  # 12
        ]
    )
    fake_data_header = pd.DataFrame(
        [
            {"CHILD": 101, "SEX": "1", "MOTHER": pd.NA},  # 0 pass: male no value
            {"CHILD": 102, "SEX": "2", "MOTHER": "0"},  # 1 pass
            {"CHILD": 103, "SEX": "2", "MOTHER": 0},  # 2 pass
            {"CHILD": 104, "SEX": "2", "MOTHER": 1},  # 3 pass
            {"CHILD": 105, "SEX": "2", "MOTHER": pd.NA},  # 4 fail: no value
            {"CHILD": 106, "SEX": "2", "MOTHER": "2"},  # 5 fail: invalid value
            {"CHILD": 107, "SEX": "1", "MOTHER": "1"},  # 6 fail: male value
            {
                "CHILD": 108,
                "SEX": "2",
                "MOTHER": pd.NA,
            },  # 7 fail: has OC3 data but also has episode in collection year
            {
                "CHILD": 109,
                "SEX": "2",
                "MOTHER": pd.NA,
            },  # 8 pass: has OC3 and no episode in collection year
            {"CHILD": 110, "SEX": "2", "MOTHER": 1},
            # 9 pass: no non-V3/V4 episode in collection year and no OC3
            {"CHILD": 111, "SEX": "2", "MOTHER": pd.NA},  # 10 pass: V3 episode
        ]
    )
    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}
    fake_dfs = {
        "Header": fake_data_header,
        "Episodes": fake_data_episodes,
        "OC3": fake_data_oc3,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {"Header": [4, 5, 6, 7], "OC3": [4, 5, 6, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="565",
    message="The date that the child started to be missing or away from placement without authorisation has been completed but whether the child was missing or away from placement without authorisation has not been completed.",
    affected_fields=["MISSING", "MIS_START"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        missing = dfs["Missing"]
        mask = missing["MIS_START"].notna() & missing["MISSING"].isna()
        error_locations = missing.index[mask]
        return {"Missing": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"MISSING": "M", "MIS_START": pd.NA},  # 0
            {"MISSING": pd.NA, "MIS_START": "07/02/2020"},  # 1
            {"MISSING": "A", "MIS_START": "03/02/2020"},  # 2
            {"MISSING": pd.NA, "MIS_START": pd.NA},  # 3
            {"MISSING": "M", "MIS_START": pd.NA},  # 4
            {"MISSING": "A", "MIS_START": "13/02/2020"},  # 5
        ]
    )

    fake_dfs = {"Missing": fake_data}

    assert validate(fake_dfs) == {"Missing": [1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="575",
    message="If the placement from which the child goes missing/away from placement without "
    + "authorisation ends, the missing/away from placement without authorisation period in "
    + "the missing module must also have an end date.",
    affected_fields=["MIS_END"],
    tables=["Episodes", "Missing"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Missing" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        mis = dfs["Missing"]

        mis.reset_index(inplace=True)

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")
        mis["MIS_START"] = pd.to_datetime(
            mis["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )

        m_coh = epi.merge(mis, how="inner", on="CHILD").query(
            "(MIS_START >= DECOM) & (MIS_START <= DEC) & MIS_END.isnull() & DEC.notnull()"
        )

        err_list = m_coh["index"].unique().tolist()
        err_list.sort()
        return {"Missing": err_list}

        return {"Episodes": Episodes_errs, "AD1": AD1_errs}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        [
            {"CHILD": "111", "DECOM": "01/06/2020", "DEC": "10/06/2020"},  # 0
            {"CHILD": "123", "DECOM": "08/06/2020", "DEC": "10/06/2020"},  # 1
            {"CHILD": "222", "DECOM": "05/06/2020", "DEC": "10/06/2020"},  # 2
            {"CHILD": "333", "DECOM": "06/06/2020", "DEC": pd.NA},  # 3
            {"CHILD": "333", "DECOM": "07/06/2020", "DEC": "10/06/2020"},  # 4
            {"CHILD": "444", "DECOM": "08/06/2020", "DEC": "10/06/2020"},  # 5
            {"CHILD": "444", "DECOM": "08/06/2020", "DEC": pd.NA},  # 6
        ]
    )

    fake_mis = pd.DataFrame(
        [
            {"CHILD": "111", "MIS_START": "01/06/2020", "MIS_END": "05/06/2020"},  # 0
            {"CHILD": "123", "MIS_START": "08/06/2020", "MIS_END": pd.NA},  # 1 Fail
            {"CHILD": "222", "MIS_START": "05/06/2020", "MIS_END": "05/06/2020"},  # 2
            {"CHILD": "333", "MIS_START": "06/06/2020", "MIS_END": pd.NA},  # 3
            {"CHILD": "444", "MIS_START": "08/06/2020", "MIS_END": "05/06/2020"},  # 4
            {"CHILD": "444", "MIS_START": "09/06/2020", "MIS_END": pd.NA},  # 5 Fail
            {"CHILD": "444", "MIS_START": "19/06/2020", "MIS_END": "05/06/2020"},  # 6
        ]
    )

    fake_dfs = {"Episodes": fake_epi, "Missing": fake_mis}

    result = validate(fake_dfs)

    assert result == {"Missing": [1, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="519",
    message="Data entered on the legal status of adopters shows civil partnership couple, but data entered on genders of adopters does not show it as a couple.",
    affected_fields=["LS_ADOPTR", "SEX_ADOPTR"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}
    else:
        ad1 = dfs["AD1"]
        mask = (ad1["LS_ADOPTR"] == "L2") & (
            (ad1["SEX_ADOPTR"] != "MM")
            & (ad1["SEX_ADOPTR"] != "FF")
            & (ad1["SEX_ADOPTR"] != "MF")
        )
        error_locations = ad1.index[mask]
        return {"AD1": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "LS_ADOPTR": ["L2", "L4", pd.NA, "L2", "L2", "L2", "L4"],
            "SEX_ADOPTR": ["MM", "FF", "MM", "M1", "MF", pd.NA, "xxxxx"],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)

    assert result == {"AD1": [3, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


import pandas as pd


@rule_definition(
    code="1010",
    message="This child has no episodes loaded for current year even though there was an open episode of "
    + "care at the end of the previous year, and care leaver data has been entered.",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["Episodes", "Episodes_last", "OC3"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Episodes_last" not in dfs or "OC3" not in dfs:
        return {}

    else:
        episodes = dfs["Episodes"]
        episodes_last = dfs["Episodes_last"]
        oc3 = dfs["OC3"]

        # convert DECOM to datetime, drop missing/invalid sort by CHILD then DECOM,
        episodes_last["DECOM"] = pd.to_datetime(
            episodes_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes_last = episodes_last.dropna(subset=["DECOM"]).sort_values(
            ["CHILD", "DECOM"], ascending=True
        )

        # Keep only the final episode for each child (ie where the following row has a different CHILD value)
        episodes_last = episodes_last[
            episodes_last["CHILD"].shift(-1) != episodes_last["CHILD"]
        ]
        # Keep only the final episodes that were still open
        episodes_last = episodes_last[episodes_last["DEC"].isna()]

        # The remaining children ought to have episode data in the current year if they are in OC3
        has_current_episodes = oc3["CHILD"].isin(episodes["CHILD"])
        has_open_episode_last = oc3["CHILD"].isin(episodes_last["CHILD"])

        error_mask = ~has_current_episodes & has_open_episode_last

        validation_error_locations = oc3.index[error_mask]

        return {"OC3": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_episodes_prev = pd.DataFrame(
        {
            "CHILD": ["101", "101", "102", "102", "103", "103", "103"],
            "DEC": [
                "20/06/2020",
                pd.NA,  # ok - has current episodes
                pd.NA,
                "14/03/2021",  # ok - 'open' episode not latest by DECOM
                "01/01/2020",
                pd.NA,
                "01/05/2020",
            ],  # Fail!
            "DECOM": [
                "01/01/2020",
                "11/11/2020",
                "03/10/2020",
                "11/11/2020",
                "01/01/2020",
                "11/11/2020",
                "01/02/2020",
            ],
        }
    )

    fake_episodes = pd.DataFrame({"CHILD": ["101"], "DEC": ["20/05/2021"]})

    fake_oc3 = pd.DataFrame({"CHILD": ["101", "102", "103"]})

    fake_dfs = {
        "Episodes": fake_episodes,
        "Episodes_last": fake_episodes_prev,
        "OC3": fake_oc3,
    }

    result = validate(fake_dfs)

    assert result == {"OC3": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="393",
    message="Child is looked after but mother field is not completed.",
    affected_fields=["MOTHER"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Header" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]

        header_female = header[header["SEX"].astype(str) == "2"]

        applicable_episodes = episodes[~episodes["LS"].str.upper().isin(["V3", "V4"])]

        error_mask = (
            header_female["CHILD"].isin(applicable_episodes["CHILD"])
            & header_female["MOTHER"].isna()
        )

        error_locations = header_female.index[error_mask]

        return {"Header": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105"],
            "SEX": ["2", "1", "2", "2", "2"],
            "MOTHER": ["1", pd.NA, "0", pd.NA, pd.NA],
        }
    )

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105"],
            "LS": ["C2", "C2", "c2", "C1", "v4"],
        }
    )

    fake_dfs = {"Header": fake_data, "Episodes": fake_data_episodes}

    result = validate(fake_dfs)

    assert result == {"Header": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="218",
    message="Ofsted Unique reference number (URN) is required.",
    affected_fields=["URN"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        collection_start = dfs["metadata"]["collection_start"]
        pl_list = [
            "H5",
            "P1",
            "P2",
            "P3",
            "R1",
            "R2",
            "R5",
            "T0",
            "T1",
            "T2",
            "T3",
            "T4",
            "Z1",
        ]

        # convert string date values to datetime format to enable comparison.
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC_dt"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        out_of_england = (
            episodes["PL_LA"]
            .astype("str")
            .str.upper()
            .str.startswith(("N", "W", "S", "C"))
        )
        place_exempt = episodes["PLACE"].isin(pl_list)
        ends_after_collection_start = (
            episodes["DEC_dt"] >= collection_start
        ) | episodes["DEC"].isna()

        mask = (
            ends_after_collection_start
            & episodes["URN"].isna()
            & ~place_exempt
            & ~out_of_england
        )

        error_locations = episodes.index[mask]
        return {"Episodes": error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "PLACE": "H5",
                "DEC": "01/02/2014",
                "PL_LA": "o",
                "URN": "xx",
            },  # 0 ignore: of PLACE value
            {
                "CHILD": "1111",
                "PLACE": "C2",
                "DEC": "01/06/2015",
                "PL_LA": "W06000008",
                "URN": "xx",
            },  # 1 ignore: PL_LA is Wales
            {
                "CHILD": "1111",
                "PLACE": "C2",
                "DEC": "01/02/2016",
                "PL_LA": "o",
                "URN": "xx",
            },  # 2 pass
            {
                "CHILD": "2222",
                "PLACE": "C2",
                "DEC": "01/02/2014",
                "PL_LA": "o",
                "URN": "xx",
            },  # 3 ignore: DEC < coll. end
            {
                "CHILD": "2222",
                "PLACE": "xx",
                "DEC": pd.NA,
                "PL_LA": "o",
                "URN": pd.NA,
            },  # 4 fail
            {
                "CHILD": "2222",
                "PLACE": "R5",
                "DEC": "01/02/2016",
                "PL_LA": "o",
                "URN": "xx",
            },  # 5 ignore: of PLACE value
            {
                "CHILD": "3333",
                "PLACE": "C2",
                "DEC": "01/03/2014",
                "PL_LA": "o",
                "URN": pd.NA,
            },  # 6 ignore: DEC
            {
                "CHILD": "3333",
                "PLACE": "C2",
                "DEC": "04/01/2016",
                "PL_LA": "o",
                "URN": pd.NA,
            },  # 7 fail
            {
                "CHILD": "4444",
                "PLACE": "C2",
                "DEC": "01/02/2016",
                "PL_LA": "o",
                "URN": pd.NA,
            },  # 8 fail
            {
                "CHILD": "4444",
                "PLACE": "C2",
                "DEC": "01/04/2017",
                "PL_LA": "o",
                "URN": pd.NA,
            },  # 9 fail
            {
                "CHILD": "5555",
                "PLACE": "C2",
                "DEC": "31/03/2015",
                "PL_LA": "o",
                "URN": pd.NA,
            },  # 10 ignore: DEC
            {
                "CHILD": "5555",
                "PLACE": "C2",
                "DEC": "04/01/2016",
                "PL_LA": "SCO",
                "URN": pd.NA,
            },  # 11 ignore: in Scotland
        ]
    )

    metadata = {"collection_start": "01/04/2015"}

    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)
    assert result == {"Episodes": [4, 7, 8, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="188",
    message="Child is aged under 4 years at the end of the year, "
    "but a Strengths and Difficulties (SDQ) score or a reason "
    "for no SDQ score has been completed. ",
    affected_fields=["SDQ_SCORE", "SDQ_REASON"],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}

    oc2 = dfs["OC2"]

    collection_end_str = dfs["metadata"]["collection_end"]

    collection_end = pd.to_datetime(
        collection_end_str, format="%d/%m/%Y", errors="coerce"
    )
    oc2["DOB_dt"] = pd.to_datetime(oc2["DOB"], format="%d/%m/%Y", errors="coerce")

    oc2["4th_bday"] = oc2["DOB_dt"] + pd.DateOffset(years=4)
    error_mask = (oc2["4th_bday"] > collection_end) & oc2[
        ["SDQ_SCORE", "SDQ_REASON"]
    ].notna().any(axis=1)

    oc2_errors = oc2.loc[error_mask].index.to_list()

    return {"OC2": oc2_errors}


def test_validate():
    import pandas as pd

    metadata = {"collection_end": "31/03/1981"}

    oc2 = pd.DataFrame(
        {
            "CHILD": ["0", "1", "2", "3", "4"],
            "DOB": [
                "01/01/1970",
                "01/01/1980",
                "01/01/1970",
                "31/03/1977",
                "01/04/1977",
            ],
            "SDQ_SCORE": ["OK", "!!!", pd.NA, "OK", "!!!"],
        }
    )

    oc2["SDQ_REASON"] = pd.NA

    fake_dfs = {"OC2": oc2, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"OC2": [1, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="213",
    message="Placement provider information not required.",
    affected_fields=["PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        mask = (
            df["PLACE"].isin(["T0", "T1", "T2", "T3", "T4"])
            & df["PLACE_PROVIDER"].notna()
        )
        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["T0", "U6", "U1", "U4", "P1", "T2", "T3", pd.NA, "Z1", "Z1"],
            "PLACE_PROVIDER": [
                pd.NA,
                pd.NA,
                "PR3",
                "PR4",
                "PR0",
                "PR2",
                "PR1",
                pd.NA,
                pd.NA,
                "notna",
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [5, 6]}



import logging
import os
from copy import copy
from pathlib import Path
from typing import Any, Dict

import numpy as np
import pandas as pd
import qlacref_authorities
from pandas import DataFrame
from qlacref_postcodes import Postcodes

logger = logging.getLogger(__name__)

# TODO security point. remove this line.
os.environ["QLACREF_PC_INSECURE"] = "True"

la_df = pd.DataFrame.from_records(qlacref_authorities.records)
logger.info("Loaded authorities")

# A bit of a hack, but will keep things working
if os.getenv("QLACREF_PC_KEY") is None:
    key_file = Path(__file__).parent.parent / ".qlacref/id_rsa.pub"
    if key_file.is_file():
        logger.warning(
            "Using repository key for pickle signature. "
            "To stay safe from tampering, set the key path in 'QLACREF_PC_KEY'"
        )
        os.environ["QLACREF_PC_KEY"] = str(key_file.absolute())

postcodes = Postcodes()
logger.info("Initialised Postcodes")


def create_datastore(data: Dict[str, Any], metadata: Dict[str, Any]):
    """
    Returns a dictionary with keys for
    - Every table name in config.py
    - Every table name again, with the suffix '_last', if last years data was uploaded.
    - A 'metadata' key, which links to a nested dictionary with
      - 'postcodes' - a postcodes csv, with columns "laua" (LA code), "oseast1m", "osnrth1m" (coordinates) and "pcd" (the postcode)
      - 'localAuthority' - the code of the local authority (long form)
      - 'collectionYear' - the collection year string (e.g. '2019/20)

    :param data: Dict of raw DataFrames by name (from config.py) together with the '_last' data.
    :param metadata:
    """
    logger.info(", ".join(data.keys()))
    data = copy(data)
    data["metadata"] = _process_metadata(metadata)
    if "Episodes" in data:
        data["Episodes"] = _add_postcode_derived_fields(
            data["Episodes"],
            metadata["localAuthority"],
        )
    if "Episodes_last" in data:
        data["Episodes_last"] = _add_postcode_derived_fields(
            data["Episodes_last"], metadata["localAuthority"]
        )
    # quick n dirty fix for weird postcode related columns showing up in episode tables
    for table_name in ["Episodes", "Episodes_last"]:
        if table_name in data:
            data[table_name] = data[table_name].drop(
                columns={"_14", "_15", "_16", "_17"} & set(data[table_name].columns)
            )

    names_and_lengths = ", ".join(f"{t}: {len(data[t])} rows" for t in data)
    logger.info(f"Datastore created -- {names_and_lengths}")
    return data


def _process_metadata(metadata):
    collection_year = int(metadata["collectionYear"][:4])
    metadata["collection_start"] = f"01/04/{collection_year - 1}"
    metadata["collection_end"] = f"31/03/{collection_year}"
    return metadata


def merge_postcodes(df: DataFrame, postcode_field: str) -> DataFrame:
    df[postcode_field] = df[postcode_field].str.upper()

    key_field = f"__{postcode_field}__first_letter"
    df[key_field] = df[postcode_field].str[0]
    pcs = df[key_field].dropna().unique()

    logger.info(f"Loading the following postcode letters: {pcs}")
    postcodes.load_postcodes(pcs)

    logger.info(f"Adding postcode abbreviations for {postcode_field}")
    pc_abbr_field = f"__{postcode_field}__abbr"
    df[pc_abbr_field] = df[postcode_field].str.replace(" ", "")

    df_merged = df[[pc_abbr_field]].merge(
        postcodes.dataframe, how="left", left_on=pc_abbr_field, right_on="pcd_abbr"
    )
    del df[key_field]
    del df[pc_abbr_field]

    return df_merged


def _add_postcode_derived_fields(episodes_df, local_authority):
    episodes_df = episodes_df.copy()

    home_details = merge_postcodes(episodes_df, "HOME_POST")
    pl_details = merge_postcodes(episodes_df, "PL_POST")

    # The indices remain the same post merge as the length of the dataframes doesn't change, so we can set directly.
    pl_details = pl_details.merge(
        la_df, how="left", left_on="laua", right_on="LTLA21CD"
    )
    episodes_df["PL_LA"] = pl_details["UTLA21CD"]

    logger.info(f"Adding IN/OUT")
    episodes_df["PL_LOCATION"] = "IN"
    episodes_df.loc[episodes_df["PL_LA"].ne(local_authority), "PL_LOCATION"] = "OUT"
    episodes_df.loc[pl_details["laua"].isna(), "PL_LOCATION"] = pd.NA

    # Add country codes for placements outside england
    for letter, code in {"S": "SCO", "N": "NIR", "W": "WAL"}.items():
        mask = (
            pl_details["laua"].str.upper().str.startswith(letter)
            & episodes_df["PL_POST"].notnull()
        )
        episodes_df.loc[mask, ["PL_LA", "PL_LOCATION"]] = [code, "OUT"]

    logger.info(f"Calculating distances")
    # This formula is taken straight from the guidance, to get miles between two postcodes
    episodes_df["PL_DISTANCE"] = (
        np.sqrt(
            (home_details["oseast1m"] - pl_details["oseast1m"]) ** 2
            + (home_details["osnrth1m"] - pl_details["osnrth1m"]) ** 2
        )
        / 1000
        / 1.6093
    )
    episodes_df["PL_DISTANCE"] = episodes_df["PL_DISTANCE"].round(decimals=1)

    logger.info(f"Completed postcode calculations")
    return episodes_df


def copy_datastore(data_store):
    """
    This is used for getting a shallow copy that is passed to the validation rules.
    """
    return {
        k: v.copy(deep=False) if k != "metadata" else v for k, v in data_store.items()
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="392a",
    message="Child is looked after but no distance is recorded. [NOTE: This check may result in false positives for children formerly UASC]",
    affected_fields=["PL_DISTANCE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        # If <LS> not = 'V3' or 'V4' and <UASC> = '0' and <COLLECTION YEAR> - 1 <UASC> = '0' and <COLLECTION YEAR> - 2 <UASC> = '0' then <PL_DISTANCE> must be provided
        epi = dfs["Episodes"]
        epi["orig_idx"] = epi.index

        header = pd.DataFrame()
        if "Header" in dfs:
            header = pd.concat((header, dfs["Header"]), axis=0)
        elif "UASC" in dfs:
            uasc = dfs["UASC"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "Header_last" in dfs:
            header = pd.concat((header, dfs["Header_last"]), axis=0)
        elif "UASC_last" in dfs:
            uasc = dfs["UASC_last"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "UASC" in header.columns:
            header = header[header.UASC == "1"].drop_duplicates("CHILD")
            epi = epi.merge(
                header[["CHILD", "UASC"]], how="left", on="CHILD", indicator=True
            )
            epi = epi[epi["_merge"] == "left_only"]
        else:
            return {}

        # Check that the episodes LS are neither V3 or V4.
        epi = epi.query("(~LS.isin(['V3','V4'])) & ( PL_DISTANCE.isna())")
        err_list = epi["orig_idx"].tolist()
        err_list.sort()

        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_episodes = pd.DataFrame(
        [
            {"CHILD": "111", "LS": "L1", "PL_DISTANCE": "XX1"},  # 0
            {"CHILD": "222", "LS": "L1", "PL_DISTANCE": pd.NA},  # 1 fail
            {"CHILD": "222", "LS": "V3", "PL_DISTANCE": "XX1"},  # 2
            {"CHILD": "333", "LS": "L1", "PL_DISTANCE": pd.NA},  # 3 fail
            {"CHILD": "333", "LS": "V4", "PL_DISTANCE": "XX1"},  # 4
            {"CHILD": "345", "LS": "L1", "PL_DISTANCE": pd.NA},  # 5 fail
            {"CHILD": "444", "LS": "L1", "PL_DISTANCE": "XX1"},  # 6
            {
                "CHILD": "444",
                "LS": "V3",
                "PL_DISTANCE": pd.NA,
            },  # 7 pass since LS isin [V3, V4]
            {"CHILD": "555", "LS": "L1", "PL_DISTANCE": pd.NA},  # 8 fail
        ]
    )
    fake_header = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "1"},  # 0
            {"CHILD": "222", "UASC": "0"},  # 2
            {"CHILD": "333", "UASC": "0"},  # 4
            {"CHILD": "345", "UASC": "0"},  # 5
            {"CHILD": "444", "UASC": "0"},  # 6
            {"CHILD": "555", "UASC": "0"},  # 6
        ]
    )
    fake_header_last = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "0"},  # 0
            {"CHILD": "222", "UASC": "0"},  # 2
            {"CHILD": "333", "UASC": "0"},  # 4
            {"CHILD": "345", "UASC": "0"},  # 5
            {"CHILD": "444", "UASC": "1"},  # 6
        ]
    )

    dfs = {
        "Episodes": fake_episodes,
        "Header": fake_header,
        "Header_last": fake_header_last,
    }
    result = validate(dfs)
    assert result == {"Episodes": [1, 3, 5, 8]}

    uasc_last = pd.DataFrame(
        [
            {"CHILD": "333", "DUC": "something"},  # [3] in episodes - passes now
            {"CHILD": "555", "DUC": pd.NA},  # [8] in episodes - still fails
        ]
    )

    dfs = {"Episodes": fake_episodes, "Header": fake_header, "UASC_last": uasc_last}
    result = validate(dfs)
    assert result == {"Episodes": [1, 5, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="383",
    message="A child in a temporary placement must subsequently return to his/her normal placement.",
    affected_fields=["PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi.sort_values(["CHILD", "DECOM"], inplace=True)

        epi.reset_index(inplace=True)
        epi.reset_index(inplace=True)
        epi["LAG_INDEX"] = epi["level_0"].shift(-1)
        epi["LEAD_INDEX"] = epi["level_0"].shift(1)

        m_epi = epi.merge(
            epi,
            how="inner",
            left_on="level_0",
            right_on="LAG_INDEX",
            suffixes=["", "_TOP"],
        )
        m_epi = m_epi.merge(
            epi,
            how="inner",
            left_on="level_0",
            right_on="LEAD_INDEX",
            suffixes=["", "_BOTM"],
        )
        m_epi = m_epi[m_epi["CHILD"] == m_epi["CHILD_TOP"]]
        m_epi = m_epi[m_epi["CHILD"] == m_epi["CHILD_BOTM"]]
        m_epi = m_epi[m_epi["PLACE"].isin(["T0", "T1", "T2", "T3", "T4"])]

        mask1 = m_epi["RNE_BOTM"] != "P"
        mask2 = m_epi["PLACE_BOTM"] != m_epi["PLACE_TOP"]
        err_mask = mask1 | mask2
        err_list = m_epi["index"][err_mask].unique().tolist()
        err_list.sort()
        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "111", "RNE": "S", "DECOM": "01/06/2020", "PLACE": "S1"},  # 0
            {
                "CHILD": "111",
                "RNE": "S",
                "DECOM": "05/06/2020",
                "PLACE": "T1",
            },  # 1 Middle Fails as next RNE not P
            {"CHILD": "111", "RNE": "T", "DECOM": "06/06/2020", "PLACE": "S1"},  # 2
            {"CHILD": "222", "RNE": "S", "DECOM": "05/06/2020", "PLACE": "T1"},  # 3
            {"CHILD": "333", "RNE": "S", "DECOM": "06/06/2020", "PLACE": "S1"},  # 4
            {
                "CHILD": "333",
                "RNE": "S",
                "DECOM": "10/06/2020",
                "PLACE": "T1",
            },  # 5 Middle Fails as pre PL not next PL
            {"CHILD": "333", "RNE": "P", "DECOM": "12/06/2020", "PLACE": "S2"},  # 6
            {"CHILD": "444", "RNE": "S", "DECOM": "08/06/2020", "PLACE": "T1"},  # 7
            {
                "CHILD": "444",
                "RNE": "S",
                "DECOM": "09/06/2020",
                "PLACE": "P4",
            },  # 8  Middle Passes not a T code
            {"CHILD": "444", "RNE": "S", "DECOM": "15/06/2020", "PLACE": "T1"},  # 9
            {"CHILD": "6666", "RNE": "S", "DECOM": "12/06/2020", "PLACE": "T1"},  # 10
            {
                "CHILD": "6666",
                "RNE": "S",
                "DECOM": "13/06/2020",
                "PLACE": "T1",
            },  # 11 Middle Passes
            {"CHILD": "6666", "RNE": "P", "DECOM": "14/06/2020", "PLACE": "T1"},  # 12
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 5]}
